package sk.fiit.jim.agent.models;

import java.util.HashMap;

import static sk.fiit.jim.agent.models.EnvironmentModel.*;
import static sk.fiit.jim.log.LogType.AGENT_MODEL;

import java.util.Map;
import java.util.Map.Entry;

import org.jruby.compiler.ir.operands.Array;

import static java.lang.Math.PI;
import static java.lang.Math.toDegrees;

import sk.fiit.jim.Settings;
import sk.fiit.jim.agent.moves.Joint;
import sk.fiit.jim.agent.parsing.ParsedData;
import sk.fiit.jim.agent.parsing.ParsedDataObserver;
import sk.fiit.jim.code_review.Problem;
import sk.fiit.jim.code_review.UnderConstruction;
import sk.fiit.jim.log.Log;
import sk.fiit.jim.math.Angles;
import sk.fiit.jim.math.Vector3D;

/**
 *  AgentModel.java
 *  
 *  Data structure encapsulating information about
 *  agent's current state, such as joint positions,
 *  global position and rotations against global axis.
 *  
 *  Calculations are delegated to {@link AgentPositionCalculator} and {@link AgentRotationCalculator}.
 *
 *@Title        Jim
 *@author       $Author: marosurbanec $
 */
public class AgentModel implements ParsedDataObserver{
	
	private static AgentModel instance = new AgentModel();
	double rotationX = 0.0;
	double rotationY = 0.0;
	double rotationZ = 3.0*Math.PI / 2.0;
	Vector3D position = Vector3D.cartesian(0, 0, 0);
	Vector3D velocity = Vector3D.cartesian(0, 0, 0); //added by Ateam - todo
	Vector3D pureBodyAcceleration = Vector3D.cartesian(0, 0, 0); //added by Ateam - zrychlenie kde nie je zapocitane gravitacne zrychlenie
	private Map<Joint, Double> jointAngles = new HashMap<Joint, Double>();
	ParsedData lastDataReceived = new ParsedData();
	
	public AgentModel(){
		for (Joint joint : Joint.values())
			jointAngles.put(joint, 0.0);
	}
	
	public static AgentModel getInstance(){
		return instance;
	}
	
	public static double getJointAngleOf(Joint joint){
		return instance.jointAngles.get(joint);
	}

	public void processNewServerMessage(ParsedData data){
		lastDataReceived = data;
		getPosition(data);
		writeAccData(data);
		//writeGyroData(data);
		updateJointPositions(data);
		adjustRotationsFor(data.gyroscope);
		//adjustPositionFor(data.accelerometer);
		updateRotations(data);
		updatePureBodyAcceleration(data);
		//updateVelocity(pureBodyAcceleration);
		updatePosition(data);
	}

	private void updateJointPositions(ParsedData data){
		for (Entry<Joint, Double> record : data.agentsJoints.entrySet())
			this.jointAngles.put(record.getKey(), record.getValue());
	}
	
	private void writeGyroData(ParsedData data){
		Log.log(AGENT_MODEL, "Gyroscope: [%.5f,%.5f,%.5f]", data.gyroscope.getX() ,data.gyroscope.getY(),data.gyroscope.getZ());
	}
	
	private void writeAccData(ParsedData data){
		Log.log(AGENT_MODEL, "Acceleration: [%.5f,%.5f,%.5f]", data.accelerometer.getX() ,data.accelerometer.getY(),data.accelerometer.getZ());
	}
	
	private Vector3D rotateVectorAroundVector(Vector3D povodny, Vector3D os, double radUhol){
		Vector3D vysledny = null;
		double c = Math.cos(radUhol);
		double s = Math.sin(radUhol);
		
		double array[][] = new double[3][3];
		
		//vytvorim transformacnu maticu ktorou vynasobim povodny vektor
		array[0][0]= (os.getX() * os.getX()) * (1.0f - c) + c;
		array[0][1]= (os.getY() * os.getX()) * (1.0f - c) + (os.getZ() * s);
		array[0][2]= (os.getZ() * os.getX()) * (1.0f - c) - (os.getY() * s);
		array[1][0]=(os.getX() * os.getY()) * (1.0f - c) - (os.getZ() * s);
		array[1][1]=(os.getY() * os.getY()) * (1.0f - c) + c;
		array[1][2]=(os.getZ() * os.getY()) * (1.0f - c) + (os.getX() * s);
		array[2][0]=(os.getX() * os.getZ()) * (1.0f - c) + (os.getY() * s);
		array[2][1]=(os.getY() * os.getZ()) * (1.0f - c) - (os.getX() * s);
		array[2][2]=(os.getZ() * os.getZ()) * (1.0f - c) + c;
		
		vysledny = vectorTimesMatrix(array,povodny);
		
		return vysledny;
	}
	
	private Vector3D rotateVectorAroundZAxis(Vector3D povodny, double uhol){
		double x = povodny.getX() * Math.cos(uhol) - povodny.getY() * Math.sin(uhol);
		double y = povodny.getX() * Math.sin(uhol) + povodny.getY() * Math.cos(uhol);
		
		Vector3D vysledny = Vector3D.cartesian(x,y,povodny.getZ());
		return vysledny;
	}
	
	//otacam X a Y okolo Z a potom X okolo uz zrotovaneho Yonu
	private Vector3D countRotatedXAxis(double uholokoloz, double uholokoloy){
		Vector3D x = Vector3D.cartesian(1, 0, 0);
		Vector3D y= Vector3D.cartesian(0, 1, 0);
		
		Vector3D XpoPrvomOtoceni = rotateVectorAroundZAxis(x,uholokoloz);//okolo osi z
		Vector3D YpoPrvomOtoceni = rotateVectorAroundZAxis(y,uholokoloz);//okolo osi z
		
		//teraz musim otocit X okolo natoceneho Y
		Vector3D vysledny = rotateVectorAroundVector( XpoPrvomOtoceni, YpoPrvomOtoceni, uholokoloy);
		
		return vysledny; // toto je Z rotovane podla X a Y
	}
	
	private Vector3D vectorTimesMatrix(double[][] matrix, Vector3D vector){
		//vynasobi vektor s maticou
		double x = vector.getX()*matrix[0][0] + vector.getX()*matrix[0][1] + vector.getX()*matrix[0][2] ;
		double y = vector.getY()*matrix[1][0] + vector.getY()*matrix[1][1] + vector.getY()*matrix[1][2] ;
		double z = vector.getZ()*matrix[2][0] + vector.getZ()*matrix[2][1] + vector.getZ()*matrix[2][2] ;
		
		Vector3D vysledny = Vector3D.cartesian(x,y,z);
		
		return vysledny;
	}
	
	private double getAngleBetweenTwoVectors(Vector3D prvy, Vector3D druhy){
		double uhol;
		uhol = Math.acos(prvy.getX() * druhy.getX() + prvy.getY() * druhy.getY() + prvy.getZ() * druhy.getZ());
		return uhol;
	}
	
	//added by Ateam 
	private void updatePureBodyAcceleration(ParsedData data){
		Vector3D oldX = Vector3D.cartesian(1, 0, 0);
		Vector3D oldY = Vector3D.cartesian(0, 1, 0);
		Vector3D oldZ = Vector3D.cartesian(0, 0, 1);
		
		Vector3D rotatedX = countRotatedXAxis(rotationZ, rotationY);
		Vector3D rotatedY = Vector3D.cartesian(rotatedX.getY(), rotatedX.getX(), rotatedX.getZ());
		Vector3D rotatedZ = Vector3D.cartesian(rotatedY.getX(), rotatedY.getZ(), rotatedY.getY());
		
		
		if ( data.accelerometer == null ) return;
		
		pureBodyAcceleration = Vector3D.cartesian( data.accelerometer.getX() -  Math.cos ( getAngleBetweenTwoVectors(oldX, rotatedX) ) * Settings.getDouble("gravityAcceleration")  ,
				data.accelerometer.getY() -  Math.cos ( getAngleBetweenTwoVectors(oldY, rotatedY) )  * Settings.getDouble("gravityAcceleration") ,
				data.accelerometer.getZ() -  Math.cos ( getAngleBetweenTwoVectors(oldZ, rotatedZ) ) * Settings.getDouble("gravityAcceleration")  );
		
		Log.log(AGENT_MODEL, "Pure body acceleration: [%.5f,%.5f,%.5f]", pureBodyAcceleration.getX() ,pureBodyAcceleration.getY(),pureBodyAcceleration.getZ());
	}
	/*
	private void updateVelocity(Vector3D pureBodyAcceleration){
		if ( pureBodyAcceleration == null ) return;
		
		velocity = Vector3D.cartesian( velocity.getX() + ( TIME_STEP * pureBodyAcceleration.getX() ) ,
				velocity.getY() + ( TIME_STEP * pureBodyAcceleration.getY()) ,
				velocity.getZ() + ( TIME_STEP * pureBodyAcceleration.getZ()) );
	
		Log.log(AGENT_MODEL, "Velocity: [%.5f,%.5f,%.5f]", velocity.getX(),velocity.getY(),velocity.getZ());
	}*/
	
	
	/*
	private void adjustPositionFor(V)
	{
		if (accelerometer == null) return;
		position = Vector3D.cartesian(
			position.getX() + accelerometer.getX()*TIME_STEP,
			position.getY() + accelerometer.getY()*TIME_STEP,
			position.getZ() + (accelerometer.getZ() - Settings.getDouble("gravityAcceleration"))*TIME_STEP
		);
		
		Log.log(AGENT_MODEL, "Accelerometer values: [%.5f,%.5f,%.5f]", accelerometer.getX(), accelerometer.getY(), accelerometer.getZ());
		//Log.log(AGENT_MODEL, "My new position: [%.5f,%.5f,%.5f]", position.getX(), position.getY(), position.getZ());
		
	}*/

	private void adjustRotationsFor(Vector3D gyroscope){
		if (gyroscope == null) return;
		rotationX += Math.toRadians(gyroscope.getX() * TIME_STEP);
		rotationX = Angles.normalize(rotationX);
		rotationY += Math.toRadians(gyroscope.getY() * TIME_STEP);
		rotationY = Angles.normalize(rotationY);
		rotationZ += Math.toRadians(gyroscope.getZ() * TIME_STEP);
		rotationZ = Angles.normalize(rotationZ);
		
		Log.log(AGENT_MODEL, "Rotation: [%.5f,%.5f,%.5f]", rotationX ,rotationY,rotationZ);
	}
	
	private void updateRotations(ParsedData data){
		new AgentRotationCalculator(this).updateRotations(data);
	}
	
	private void updatePosition(ParsedData data){
		new AgentPositionCalculator(this).updatePosition(data);
	}
	
	private void getPosition(ParsedData data){
		if ( isStanding ())
			System.out.println("stojim");
		else if ( isOnGround()){
			System.out.println("na zemi");
			if ( isLyingOnBack())
				System.out.println("chrbat");
			else if (isLyingOnBelly())
				System.out.println("brucho");
		}
	}
	//---------------------QUERIES, GETTERS AND SETTERS-------------------
	public boolean isStanding(){
		return lastDataReceived.accelerometer.getZ() > 7.0;
	}
	
	public boolean isStanding(ParsedData data){
		return data.accelerometer.getZ() > 7.0;
	}
	
	public boolean isOnGround(){
		return Math.abs(lastDataReceived.accelerometer.getY()) > 9.3; 
	}
	
	public boolean isLyingOnBack(){
		return isOnGround() && ( lastDataReceived.accelerometer.getY() > 9.3 ) ;
	}
	
	public boolean isLyingOnBelly(){
		return isOnGround() && !isLyingOnBack();
	}
	
	public double getRotationX(){
		return rotationX;
	}

	public double getRotationY(){
		return rotationY;
	}

	public double getRotationZ(){
		return rotationZ;
	}

	public Vector3D getPosition(){
		return position;
	}
	
	public void setPosition(Vector3D position){
		this.position = position;
	}

	public void setRotationX(double rotationX){
		this.rotationX = rotationX;
	}

	public void setRotationY(double rotationY){
		this.rotationY = rotationY;
	}

	public void setRotationZ(double rotationZ){
		this.rotationZ = rotationZ;
	}
}